#include "Garmin.h"
#include "MacInterface.h"
#include "Prefs.h"

extern BYTE gGarminMessage[MAX_LENGTH];

static void parseGPS(void);
static void doWaypoint(void);
static void doTrack(void);
static void doRouteName(void);
static void doAlmanac(void);
static long number(BYTE *p);
static char *string(short n, BYTE *p);
static void write(char *data);

static short records;
static short FileRefNum;
static char fileData[MAX_LINE];

/* 
   The following format is used to save information about the data
   in the data files. If changed, the function getFileData() in
   SendGPSInfo.c must also be changed.
   
Format: DDD  UTC Offset:  -6.00 hrs  Datum[061]: NAD27 CONUS
        |||              ||||||            |||
0         1         2         3         4         5         6
0123456789012345678901234567890123456789012345678901234567890123456789
*/

/* --------------------------------------------------------------------------------- */

void saveFormat(char *fileData)
{
	extern struct PREFS gPrefs;
	extern struct DATUM const gDatum[];
	char format[4];

	switch (gPrefs.format) {
		case DMS: strcpy(format, "DMS"); break;
		case DMM: strcpy(format, "DMM"); break;
		case DDD: strcpy(format, "DDD"); break;
		case UTM: strcpy(format, "UTM"); break;
		case BNG: strcpy(format, "BNG"); break;
		case ITM: strcpy(format, "ITM"); break;
		case SUI: strcpy(format, "SUI"); break;
		default:  strcpy(format, "UNK"); break;
	}

	sprintf(fileData, "Format: %s  UTC Offset: %6.2f hrs  Datum[%03d]: %s\r",
		format, gPrefs.offset, gPrefs.datum, gDatum[gPrefs.datum].name);
}

/* --------------------------------------------------------------------------------- */

void getGPSInfo(short refNum, short type)
{
	extern struct PREFS gPrefs;
	extern char gMessageStr[];
	
	short done = 0;
	BYTE *init, *req;
	short total;
	BYTE *t = (BYTE *)&total;
	short err;
	char *rType;
	
	if ((err = serialOpen(GARMIN)) != noErr) {
		sprintf(gMessageStr, "The serial port initialization has failed [%d].", err);
		Error(gMessageStr);
		ExitToShell();
	}

	FileRefNum = refNum;
	records = 0;
	
	saveFormat(fileData);
	write(fileData);

	switch (type) {
		case ALMANAC:
		default:
			init = alm1;
			req = alm2;
			rType = "almanac";
			break;
			
		case ROUTE:
			init = rte1;
			req = rte2;
			rType = "route/waypoint";
			break;
			
		case TRACK:
			init = trk1;
			req = trk2;
			rType = "track";
			break;
			
		case WAYPOINT:
			init = wpt1;
			req = wpt2;
			rType = "waypoint";
			break;
	}

	InitBarGraph();
	
	sendGPSMessage(init, 4);
	if (getGPSMessage()==0) {
		CloseBarGraph();
		serialClose();
		return;
	}
	if (getGPSMessage()==0) {
		CloseBarGraph();
		serialClose();
		return;
	}
	*t++ = gGarminMessage[4];		/* MSB of number of records to transfer */
	*t = gGarminMessage[3];			/* LSB */

	while (!done) {
		sendGPSMessage(req, 4);
		if (getGPSMessage()==0)
			break;
		done = (gGarminMessage[1] == 0x0c);
		if (!done) {
			parseGPS();
			SetBarGraph((double)(++records)/(double)total);
		}
	}
	CloseBarGraph();
	serialClose();
	sprintf(gMessageStr, "%d %s packets were transferred from the GPS receiver.", total, rType);
	Message(gMessageStr);
}

/* --------------------------------------------------------------------------------- */

static void parseGPS()
{
	switch (gGarminMessage[1]) {

		case RTE_NAM:	/* Route name record */
			doRouteName();
			break;

		case RTE_WPT:	/* Route waypoint record */
			doWaypoint();
			break;

		case ALM:		/* Almanac record */
			doAlmanac();
			break;

		case TRK:		/* Track record */
			doTrack();
			break;

		case WPT:		/* Waypoint record */
			doWaypoint();
			break;

		default:
			break;
	}
}

/* --------------------------------------------------------------------------------- */

static void doRouteName()
{
	sprintf(fileData, "R\t%d\t%s\r", *(gGarminMessage+3), string(20, gGarminMessage+4));
	write(fileData);
}

/* --------------------------------------------------------------------------------- */

static void doTrack()
{
	extern struct PREFS gPrefs;
	double latitude, longitude, x, y;
	char zone[6];
	
	if (*(gGarminMessage+15) == '\x01')		/* New track, create blank record */
		if (records != 0)
			write("\r");

	latitude = int2deg(number(gGarminMessage+3));
	longitude = int2deg(number(gGarminMessage+7));

	translate(1, &latitude, &longitude, gPrefs.datum);

	sprintf(fileData, "T\t%s\t", secs2dt(number(gGarminMessage+11), gPrefs.offset));
	write(fileData);
	
	switch(gPrefs.format) {
		case DMS:
			sprintf(fileData, "%s\t", toDMS(latitude));
			write(fileData);
			sprintf(fileData, "%s\r", toDMS(longitude));
			write(fileData);
			break;

		case DMM:
			sprintf(fileData, "%s\t", toDM(latitude));
			write(fileData);
			sprintf(fileData, "%s\r", toDM(longitude));
			write(fileData);
			break;

		case DDD:
			sprintf(fileData, "%03.7f\t%04.7f\r", latitude, longitude);
			write(fileData);
			break;

		case UTM:
			DegToUTM(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%07.0f\t%07.0f\r", zone, x, y);
			write(fileData);
			break;

		case BNG:
			DegToBNG(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%05.0f\t%05.0f\r", zone, x, y);
			write(fileData);
			break;
			
		case ITM:
			DegToITM(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%05.0f\t%05.0f\r", zone, x, y);
			write(fileData);
			break;
			
		case SUI:
			break;
			
		default:
			break;
	}
}

/* --------------------------------------------------------------------------------- */

static void doWaypoint()
{
	extern struct PREFS gPrefs;
	double latitude, longitude, x, y;
	char zone[6];

	latitude = int2deg(number(gGarminMessage+9));
	longitude = int2deg(number(gGarminMessage+13));

	translate(1, &latitude, &longitude, gPrefs.datum);

	sprintf(fileData, "W\t%s\t", string(6, gGarminMessage+3));	/* W	WPT NAME */
	write(fileData);
	sprintf(fileData, "%s\t", string(40, gGarminMessage+21));	/* Comment, up to 40 characters */
	write(fileData);
	sprintf(fileData, "%s\t", secs2dt(number(gGarminMessage+17), gPrefs.offset));	/* Date/Time */
	write(fileData);

	switch(gPrefs.format) {
		case DMS:
			sprintf(fileData, "%s\t", toDMS(latitude));
			write(fileData);
			sprintf(fileData, "%s\r", toDMS(longitude));
			write(fileData);
			break;

		case DMM:
			sprintf(fileData, "%s\t", toDM(latitude));
			write(fileData);
			sprintf(fileData, "%s\r", toDM(longitude));
			write(fileData);
			break;

		case DDD:
			sprintf(fileData, "%03.7f\t%04.7f\r", latitude, longitude);
			write(fileData);
			break;

		case UTM:
			DegToUTM(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%07.0f\t%07.0f\r", zone, x, y);
			write(fileData);
			break;

		case BNG:
			DegToBNG(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%05.0f\t%05.0f\r", zone, x, y);
			write(fileData);
			break;
			
		case ITM:
			DegToITM(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%05.0f\t%05.0f\r", zone, x, y);
			write(fileData);
			break;
			
		case SUI:
			break;
			
		default:
			break;
	}
}

/* --------------------------------------------------------------------------------- */

static void doAlmanac()
{
	BYTE n, *p = gGarminMessage+3;
	
	sprintf(fileData, "A\t");
	write(fileData);
	n = gGarminMessage[2];
	while (n--) {
		sprintf(fileData, "%02X ", *p++);
		write(fileData);
	}
	write("\r");
}

/* --------------------------------------------------------------------------------- */

static char *string(short n, BYTE *p)
{
	static char s[50];
	short i;
	
	for (i=0; i<n; i++)
		s[i] = *p++;
	s[n] = '\0';
	return s;
}

/* --------------------------------------------------------------------------------- */

static long number(BYTE *p)
{	
	long n;
	BYTE *q = (BYTE *)&n;
	
	*q++ = *(p+3);
	*q++ = *(p+2);
	*q++ = *(p+1);
	*q = *p;
	
	return n;
}

static void write(char *data)
{
	long N = (long)strlen(data);
	FSWrite(FileRefNum, &N, data);
}
