#include "NMEA.h"
#include "Garmin.h"
#include "MacInterface.h"
#include "Prefs.h"

struct OUTDATA gNMEAdata;
short gDoingNMEA;

static const long TIMEOUT = 5L;	/* maximum time to wait for NMEA line */

static void outputData(void);
static void processMessage(void);
static void write(char *);
static void display(DialogPtr theDialog, short theItem, char *s);

static short FileRefNum;
static char fileData[100];
static DialogPtr NMEAwindow;

static short state;

/* The state flag is used as follows. The order of the records in the Garmin GPS 45
	NMEA data stream is GPRMC, GPRMB, GPGGA, GPGSA, GPGSV(2), PGRME, GPGLL, PGRMZ,
	PGRMM and GPRTE. Use of the state flag will force the program to output one record
	after receiving one GPRMC record, one GPGGA record and one GPGSA record. All other
	records are presently ignored, but may be easily added in the future.
	
	state	description
	  0		Ready to receive GPRMC record. Set to 1 on receipt.
	  1		Ready to receive GPGGA record. Set to 2 on receipt.
	  2		Ready to receive GPGSA record. Set to 3 on receipt.
	  3		Ready to display data. Set to 0 afterwards.
*/

/* --------------------------------------------------------------------------------- */


void doNMEA(short refNum)
{
	extern struct PREFS gPrefs;
	extern char gMessageStr[];
	
	OSErr	err;
	short windowID;
	
	FileRefNum = refNum;
	state = 0;


	if (gPrefs.format == UTM) {
		strcpy(fileData, "DATE-TIME\tZONE\t\tEASTING\tNORTHING\tALT\tVEL\tTRK\tSATS\tMODE\tPDOP\tHDOP\tVDOP\r");
		windowID = NMEA_UTM_DLOG;
	}
	else {
		strcpy(fileData, "DATE-TIME\tLATITUDE\tLONGITUDE\tALT\tVEL\tTRK\tSATS\tMODE\tPDOP\tHDOP\tVDOP\r");
		windowID = NMEA_LL_DLOG;
	}
	write(fileData);
	
	if ((err = serialOpen(NMEA)) != noErr) {
		sprintf(gMessageStr, "The serial port initialization has failed [%d].", err);
		Error(gMessageStr);
		ExitToShell();
	}
	
	else {
		NMEAwindow = GetNewDialog( windowID, nil, (WindowPtr)-1);
		SetPort(NMEAwindow);
		DrawDialog(NMEAwindow);
		ShowWindow(NMEAwindow);
		gDoingNMEA = 1;
		
		while (gDoingNMEA) {
			processMessage();
			if (state == 0)
				HandleEvent();
		}
		DisposDialog(NMEAwindow);
		serialClose();
	}
}

/* --------------------------------------------------------------------------------- */

static void processMessage()
{
	extern short gInRefNum;
	extern enum PROTOCOL mode;
	long count;
	short done = 0;
	char ch, sentence[100], *p = sentence;
	long time = TickCount();
	
	while (!done) {
		if (TickCount() > time + TIMEOUT * 60L) {
			gDoingNMEA = 0;
			Error("No NMEA output from GPS receiver.");
			mode = NONE;
			return;
		}

		while (serialCharsAvail()) {
			count = 1L;
			FSRead(gInRefNum, &count, &ch);
			if (ch == '$') {
				p = sentence;
				*p = '\0';
			}
			else {
				*p++ = ch;
				*p = '\0';
				if (ch == 0x0a)
					done = 1;
			}
		}
	}
	
	switch(state) {
	
		case 0:
		default:
			if (strncmp(GPRMC, sentence, 5) == 0)
				if (checksum(sentence)) {
					processGPRMC(sentence);
					state = 1;
				}
			break;

		case 1:
			if (strncmp(GPGGA, sentence, 5) == 0) {
				if (checksum(sentence)) {
					processGPGGA(sentence);
					state = 2;
				}
				else
					state = 0;
			}
			break;

		case 2:
			if (strncmp(GPGSA, sentence, 5) == 0) {
				if (checksum(sentence)) {
					processGPGSA(sentence);
					state = 3;
				}
				else
					state = 0;
			}
			break;

		case 3:
			outputData();
			state = 0;
			break;
	}
	return;
}

/* --------------------------------------------------------------------------------- */

void outputData()
{
	extern struct PREFS gPrefs;
	
	double latitude, longitude, x, y;
	char zone[6];
	
	latitude = gNMEAdata.latitude;
	longitude = gNMEAdata.longitude;
	
	sprintf(fileData, "%s", secs2dt(dt2secs(gNMEAdata.utc, gPrefs.offset), 0));
	write(fileData);
	write("\t");
	display(NMEAwindow, NMEA_DAT, fileData);
		
	switch(gPrefs.format) {
		case DMS:
			sprintf(fileData, "%s", toDMS(latitude));
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LAT, fileData);
			
			sprintf(fileData, "%s", toDMS(longitude));
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LON, fileData);

			break;

		case DMM:
			sprintf(fileData, "%s\t", toDM(latitude));
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LAT, fileData);
			
			sprintf(fileData, "%s\t", toDM(longitude));
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LON, fileData);

			break;

		case BNG:	// just do decimal degrees until these grids are implemented
		case ITM:
		case SUI:
		
		case DDD:
			sprintf(fileData, "%03.7f", latitude);
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LAT, fileData);
			
			sprintf(fileData, "%04.7f", longitude);
			write(fileData);
			write("\t");
			display(NMEAwindow, NMEA_LON, fileData);

			break;

		case UTM:
			DegToUTM(latitude, longitude, zone, &x, &y);
			sprintf(fileData, "%s\t%07.0f\t%07.0f\t", zone, x, y);
			write(fileData);
			
			sprintf(fileData, "%c%c%c", zone[0], zone[1], zone[3]);
			display(NMEAwindow, NMEA_UTM, fileData);
			sprintf(fileData, "%07.0f", x);
			display(NMEAwindow, NMEA_EAST, fileData);
			sprintf(fileData, "%07.0f", y);
			display(NMEAwindow, NMEA_NORTH, fileData);
			
			break;

		default:
			break;
	}

	sprintf(fileData, "%.0f\t", gNMEAdata.altitude);
	write(fileData);
	sprintf(fileData, "%.0f m (%.0f ft)", gNMEAdata.altitude,
		gNMEAdata.altitude / .3048);
	display(NMEAwindow, NMEA_ALT, fileData);

	sprintf(fileData, "%.1f\t", 1.852 * gNMEAdata.speed);
	write(fileData);
	sprintf(fileData, "%.1f km/hr (%.1f mi/hr)", 1.852 * gNMEAdata.speed,
		1.150779 * gNMEAdata.speed);
	display(NMEAwindow, NMEA_VEL, fileData);

	sprintf(fileData, "%.1f\t", gNMEAdata.track);
	write(fileData);
	sprintf(fileData, "%.1f True", gNMEAdata.track);
	display(NMEAwindow, NMEA_TRK, fileData);

	sprintf(fileData, "%d", gNMEAdata.satellites);
	write(fileData);
	write("\t");
	display(NMEAwindow, NMEA_SAT, fileData);

	sprintf(fileData, "%d\t", gNMEAdata.mode);
	write(fileData);
	switch (gNMEAdata.mode) {
		default:
		case 1:
			strcpy(fileData, "No fix");
		case 2:
			strcpy(fileData, "2D fix");
		case 3:
			strcpy(fileData, "3D fix");
	}
	display(NMEAwindow, NMEA_MODE, fileData);

	sprintf(fileData, "%.1f", gNMEAdata.pdop);
	write(fileData);
	write("\t");
	display(NMEAwindow, NMEA_PDOP, fileData);

	sprintf(fileData, "%.1f", gNMEAdata.hdop);
	write(fileData);
	write("\t");
	display(NMEAwindow, NMEA_HDOP, fileData);

	sprintf(fileData, "%.1f", gNMEAdata.vdop);
	write(fileData);
	write("\r");
	display(NMEAwindow, NMEA_VDOP, fileData);
}

/* --------------------------------------------------------------------------------- */

static void write(char *data)
{
	long N = (long)strlen(data);
	FSWrite(FileRefNum, &N, data);
}

/* --------------------------------------------------------------------------------- */

static void display(DialogPtr theDialog, short theItem, char *s)
{
	short		type;
	Handle	theHandle;
	Rect		iRect;
	
	GetDItem(theDialog, theItem, &type, &theHandle, &iRect);
	SetIText(theHandle, CtoPstr(s));
}
